import os
import re
import shutil
from pathlib import Path
from typing import Dict

from robot.parsing.model.blocks import File

from .config import Configuration
from .log import logger
from .utils import directory_to_zip


def write_test_suites(test_suites: Dict[str, File], config: Configuration) -> None:
    generation_directory = get_generation_directory(config.generationDirectory)
    if config.clearGenerationDirectory:
        clear_generation_directory(generation_directory)
    write_test_suite_files(test_suites, generation_directory)
    if config.createOutputZip:
        directory_to_zip(generation_directory)
    logger.info(f"Successfully wrote {len(test_suites)} robot files.")
    logger.info(f"Path: {os.path.abspath(generation_directory)}")


def get_generation_directory(generation_directory: str) -> Path:
    root_path = Path(os.curdir).absolute()
    if not generation_directory:
        return root_path / "Generated"
    return Path(
        re.sub(
            r"^{root}",
            str(root_path).replace('\\', '\\\\'),
            generation_directory,
            flags=re.IGNORECASE,
        )
    )


def clear_generation_directory(generation_dir: Path) -> None:
    if generation_dir.is_dir():
        shutil.rmtree(str(generation_dir))
        logger.info("Files in generation directory deleted.")
    zip_file = "".join([str(generation_dir), ".zip"])
    if os.path.exists(zip_file):
        os.remove(zip_file)


def write_test_suite_files(test_suites: Dict[str, File], generation_directory: Path) -> None:
    for test_suite_file in test_suites.values():
        test_suite_file.source = Path(generation_directory / f"{test_suite_file.source}.robot")
        logger.debug(f"File written to {os.path.relpath(test_suite_file.source)}")
        test_suite_file.save()



# pylint: skip-file
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from optparse import Option
from typing import List, Optional


class StrEnum(str, Enum):
    def __new__(cls, *args):
        for arg in args:
            if not isinstance(arg, (str, auto)):
                raise TypeError(f"Values of StrEnums must be strings: {repr(arg)} is a {type(arg)}")
        return super().__new__(cls, *args)

    def __str__(self):
        return self.value

    def _generate_next_value_(name, *_):
        return name


class FilterType(StrEnum):
    TestThemeFilter = "TestThemeFilter"
    TestCaseSetFilter = "TestCaseSetFilter"
    TestCaseFilter = "TestCaseFilter"


class TestStructureTreeNodeType(StrEnum):
    Root = "Root"
    TestTheme = "TestTheme"
    TestCaseSet = "TestCaseSet"


class Priority(StrEnum):
    High = "High"
    Medium = "Medium"
    Low = "Low"


class ReferenceType(StrEnum):
    Reference = "Reference"
    Hyperlink = "Hyperlink"
    Attachment = "Attachment"


class SpecificationStatus(StrEnum):
    NotPlanned = "NotPlanned"
    Planned = "Planned"
    InProgress = "InProgress"
    InReview = "InReview"
    Released = "Released"


class InteractionVerdict(StrEnum):
    Pass = "Pass"
    Fail = "Fail"
    Skipped = "Skipped"
    ToVerify = "ToVerify"
    Warn = "Warn"
    Undefined = "Undefined"
    Blocked = "Blocked"


class ExecutionVerdict(StrEnum):
    Undefined = "Undefined"
    Pass = "Pass"
    Fail = "Fail"
    ToVerify = "ToVerify"


class ActivityStatus(StrEnum):
    NotPlanned = "NotPlanned"
    Planned = "Planned"
    Assigned = "Assigned"
    Running = "Running"
    Canceled = "Canceled"
    Skipped = "Skipped"
    Performed = "Performed"


class ExecutionStatus(StrEnum):
    NotBlocked = "NotBlocked"
    Blocked = "Blocked"


class UdfType(StrEnum):
    String = "String"
    Enumeration = "Enumeration"
    Boolean = "Boolean"


class SequencePhase(StrEnum):
    Setup = "Setup"
    TestStep = "TestStep"
    Teardown = "Teardown"


class CallType(StrEnum):
    Check = "Check"
    Flow = "Flow"


class InteractionType(StrEnum):
    Compound = "Compound"
    Atomic = "Atomic"
    Textuell = "Textuell"


class ParameterType(StrEnum):
    DetailedInstance = "DetailedInstance"
    Unknown = "Unknown"
    InstanceTable = "InstanceTable"
    AtomicInstance = "AtomicInstance"


class ParameterUseType(StrEnum):
    CallByReference = "CallByReference"
    CallByValue = "CallByValue"
    CallByReferenceMandatory = "CallByReferenceMandatory"


@dataclass
class ProjectMember:
    userkey: str
    userLogin: str
    userName: str
    projectkey: str
    projectName: str
    roles: List[str]


@dataclass
class ProjectDetails:
    key: str
    creationTime: str
    name: str
    status: str
    visibility: bool
    tovsCount: int
    cyclesCount: int
    description: str
    lockerKey: Optional[int] = None
    startDate: Optional[str] = None
    endDate: Optional[str] = None


@dataclass
class TOVDetails:
    key: str
    creationTime: str
    name: str
    status: str
    visibility: bool
    cyclesCount: int
    description: str
    lockerKey: Optional[int] = None
    startDate: Optional[str] = None
    endDate: Optional[str] = None


@dataclass
class CycleDetails:
    key: str
    creationTime: str
    name: str
    status: str
    visibility: bool
    description: str
    startDate: Optional[str] = None
    endDate: Optional[str] = None


@dataclass
class UserDetails:
    key: str
    login: str
    name: str
    email: str
    passwordExpired: bool
    active: bool


@dataclass
class UserSummary:
    key: str
    login: str
    name: str
    active: bool

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            key=dictionary.get("key", ""),
            login=dictionary.get("login", ""),
            name=dictionary.get("name", ""),
            active=dictionary.get("active", True),
        )


@dataclass
class UserDefinedField:
    key: str
    name: str
    value: str
    valueType: UdfType

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            key=dictionary.get("key", ""),
            name=dictionary.get("name", ""),
            value=dictionary.get("value", ""),
            valueType=UdfType(dictionary.get("valueType", UdfType.String)),
        )

    @property
    def robot_tag(self) -> Optional[str]:
        if self.valueType == UdfType.Enumeration:
            return f"{self.name}:{self.value}"
        elif self.valueType == UdfType.String and self.value:
            return f"{self.name}:{self.value}"
        elif self.valueType == UdfType.Boolean and self.value == "true":
            return self.name
        return None


@dataclass
class Keyword:
    key: str
    name: str
    isVariantsMarker: bool

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            key=dictionary.get("key", ""),
            name=dictionary.get("name", ""),
            isVariantsMarker=dictionary.get("isVariantsMarker", False),
        )


@dataclass
class Reference:
    type: ReferenceType
    path: str

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            type=ReferenceType(dictionary.get("type", ReferenceType.Reference)),
            path=dictionary.get("path", ""),
        )


@dataclass
class UserReference:
    key: str
    name: str

    @classmethod
    def from_dict(cls, dictionary):
        return cls(key=dictionary.get("key", ""), name=dictionary.get("name", ""))


@dataclass
class RequirementReference:
    key: str
    edited: bool

    @classmethod
    def from_dict(cls, dictionary):
        return cls(key=dictionary.get("key", ""), edited=dictionary.get("edited", False))


@dataclass
class ConditionSummary:
    key: str
    uniqueId: str
    name: str
    description: str
    version: Optional[str] = None

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            key=dictionary.get("key", ""),
            uniqueId=dictionary.get("uniqueId", ""),
            name=dictionary.get("name", ""),
            description=dictionary.get("description", ""),
            version=dictionary.get("version", None),
        )


@dataclass
class TestCaseSetSpecificationSummary:
    key: str
    description: str
    reviewComment: str
    status: SpecificationStatus
    priority: Optional[Priority]
    responsible: Optional[UserReference]
    dueDate: Optional[str]
    reviewer: Optional[UserReference]
    udfs: List[UserDefinedField]
    keywords: List[Keyword]
    references: List[Reference]
    requirements: List[RequirementReference]
    preConditions: List[ConditionSummary]
    postConditions: List[ConditionSummary]

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            key=dictionary.get("key", ""),
            description=dictionary.get("description", ""),
            reviewComment=dictionary.get("reviewComment", ""),
            status=SpecificationStatus(dictionary.get("status", SpecificationStatus.Planned)),
            priority=Priority(dictionary.get("priority")) if dictionary.get("priority") else None,
            responsible=UserReference.from_dict(dictionary.get("responsible", {}))
            if dictionary.get("responsible")
            else None,
            dueDate=dictionary.get("dueDate", None),
            reviewer=UserReference.from_dict(dictionary.get("reviewer", {}))
            if dictionary.get("reviewer")
            else None,
            udfs=[UserDefinedField.from_dict(udf) for udf in dictionary.get("udfs", [])],
            keywords=[Keyword.from_dict(keyword) for keyword in dictionary.get("keywords", [])],
            references=dictionary.get("references", ""),
            requirements=[
                RequirementReference.from_dict(requirement)
                for requirement in dictionary.get("requirements", [])
            ],
            preConditions=[
                ConditionSummary.from_dict(condition)
                for condition in dictionary.get("preConditions", [])
            ],
            postConditions=[
                ConditionSummary.from_dict(condition)
                for condition in dictionary.get("postConditions", [])
            ],
        )


@dataclass
class TestCaseSpecificationDetails:
    key: str
    comments: str
    udfs: List[UserDefinedField]
    keywords: List[Keyword]
    requirements: List[RequirementReference]

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            key=dictionary.get("key", ""),
            comments=dictionary.get("comments", ""),
            udfs=[UserDefinedField.from_dict(udf) for udf in dictionary.get("udfs", [])],
            keywords=[Keyword.from_dict(keyword) for keyword in dictionary.get("keywords", [])],
            requirements=[
                RequirementReference.from_dict(requirement)
                for requirement in dictionary.get("requirements", [])
            ],
        )


@dataclass
class TestCaseSetExecutionSummary:
    key: str
    comments: str
    udfs: List[UserDefinedField]
    keywords: List[Keyword]

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            key=dictionary.get("key", ""),
            comments=dictionary.get("comments", ""),
            udfs=[UserDefinedField.from_dict(udf) for udf in dictionary.get("udfs", [])],
            keywords=[Keyword.from_dict(keyword) for keyword in dictionary.get("keywords", [])],
        )


@dataclass
class TestCaseSpecificationSummary:
    key: str
    comments: str
    requirements: List[RequirementReference]

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            key=dictionary.get("key", ""),
            comments=dictionary.get("comments", ""),
            requirements=[
                RequirementReference.from_dict(requirement)
                for requirement in dictionary.get("requirements", [])
            ],
        )


@dataclass
class TestCaseExecutionSummary:
    key: str
    comments: str = ""
    status: ActivityStatus = ActivityStatus.Planned
    execStatus: ExecutionStatus = ExecutionStatus.NotBlocked
    verdict: ExecutionVerdict = ExecutionVerdict.Undefined
    defects: Optional[List[int]] = None
    tester: Optional[UserReference] = None

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            key=dictionary.get("key", ""),
            status=ActivityStatus(dictionary.get("status", ActivityStatus.Planned)),
            execStatus=ExecutionStatus(dictionary.get("execStatus", ExecutionStatus.NotBlocked)),
            verdict=ExecutionVerdict(dictionary.get("verdict", ExecutionVerdict.Undefined)),
            comments=dictionary.get("comments", ""),
            defects=dictionary.get("defects", []),
            tester=UserReference.from_dict(dictionary.get("tester"))
            if dictionary.get("tester")
            else None,
        )


@dataclass
class TestCaseSummary:
    uniqueID: str
    index: int
    spec: TestCaseSpecificationSummary
    exec: Optional[TestCaseExecutionSummary] = None

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            uniqueID=dictionary.get("uniqueID", ""),
            index=dictionary.get("index", 0),
            spec=TestCaseSpecificationSummary.from_dict(dictionary.get("spec", {})),
            exec=TestCaseExecutionSummary.from_dict(dictionary.get("exec"))
            if dictionary.get("exec")
            else None,
        )


@dataclass
class TestCaseExecutionDetails:
    key: str
    status: ActivityStatus
    execStatus: ExecutionStatus
    verdict: ExecutionVerdict
    plannedDuration: int
    actualDuration: int
    currentUser: UserReference
    tester: Optional[UserReference]
    comments: str  # TODO: Insert htmlComment
    version: Optional[str]
    defects: List[int]
    udfs: List[UserDefinedField]
    keywords: List[Keyword]
    references: List[Reference]

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            key=dictionary.get("key", ""),
            status=ActivityStatus(dictionary.get("status", ActivityStatus.Planned)),
            execStatus=ExecutionStatus(dictionary.get("execStatus", ExecutionStatus.NotBlocked)),
            verdict=ExecutionVerdict(dictionary.get("verdict", ExecutionVerdict.Undefined)),
            plannedDuration=dictionary.get("plannedDuration", 0),
            actualDuration=dictionary.get("actualDuration", 0),
            currentUser=UserReference.from_dict(dictionary.get("currentUser", {})),
            tester=UserReference.from_dict(dictionary.get("tester", {}))
            if dictionary.get("tester")
            else None,
            comments=dictionary.get("comments", ""),
            version=dictionary.get("version", None),
            defects=dictionary.get("defects", []),
            udfs=[UserDefinedField.from_dict(udf) for udf in dictionary.get("udfs", [])],
            keywords=[Keyword.from_dict(keyword) for keyword in dictionary.get("keywords", [])],
            references=dictionary.get("references", []),
        )


@dataclass
class TestCaseSetDetails:
    key: str
    numbering: str
    uniqueID: str
    name: str
    spec: TestCaseSetSpecificationSummary
    exec: Optional[TestCaseSetExecutionSummary]
    testCases: List[TestCaseSummary]

    @classmethod
    def from_dict(cls, dictionary) -> TestCaseSetDetails:
        return cls(
            key=dictionary.get("key", ""),
            numbering=dictionary.get("numbering", ""),
            uniqueID=dictionary.get("uniqueID", ""),
            name=dictionary.get("name", ""),
            spec=TestCaseSetSpecificationSummary.from_dict(dictionary.get("spec", {})),
            exec=TestCaseSetExecutionSummary.from_dict(dictionary.get("exec", {}))
            if dictionary.get("exec")
            else None,
            testCases=[
                TestCaseSummary.from_dict(test_case)
                for test_case in dictionary.get("testCases", [])
            ],
        )


@dataclass
class InteractionExecutionSummary:
    verdict: InteractionVerdict
    time: str
    duration: int
    currentUser: UserReference
    tester: UserReference
    comments: str
    references: List[Reference]

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            verdict=InteractionVerdict(dictionary.get("verdict", InteractionVerdict.Undefined)),
            time=dictionary.get("time", ""),
            duration=dictionary.get("duration", 0),
            currentUser=UserReference.from_dict(dictionary.get("currentUser", {})),
            tester=UserReference.from_dict(dictionary.get("tester", {})),
            comments=dictionary.get("comments", ""),
            references=dictionary.get("references", []),
        )


@dataclass
class InteractionSpecificationSummary:
    callId: int
    sequencePhase: SequencePhase
    callType: CallType
    description: str
    comments: str
    references: List[Reference]
    preConditions: List[ConditionSummary]
    postConditions: List[ConditionSummary]

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            callId=dictionary.get("callId", ""),
            sequencePhase=SequencePhase(dictionary.get("sequencePhase", SequencePhase.TestStep)),
            callType=CallType(dictionary.get("callType", CallType.Flow)),
            description=dictionary.get("description", ""),
            comments=dictionary.get("comments", ""),
            references=dictionary.get("references", ""),
            preConditions=[
                ConditionSummary.from_dict(condition)
                for condition in dictionary.get("preConditions", [])
            ],
            postConditions=[
                ConditionSummary.from_dict(condition)
                for condition in dictionary.get("postConditions", [])
            ],
        )


@dataclass
class ParameterSummary:
    name: str
    value: str
    version: Optional[str]
    parameterType: ParameterType
    parameterUseType: ParameterUseType
    dataTypeName: str
    dataTypePath: str
    dataTypeUniqueID: str

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            name=dictionary.get("name", ""),
            value=dictionary.get("value", ""),
            version=dictionary.get(
                "version",
            ),
            parameterType=ParameterType(
                dictionary.get("parameterType", ParameterType.AtomicInstance)
            ),
            parameterUseType=ParameterUseType(
                dictionary.get("parameterUseType", ParameterUseType.CallByValue)
            ),
            dataTypeName=dictionary.get("dataTypeName", ""),
            dataTypePath=dictionary.get("dataTypePath", ""),
            dataTypeUniqueID=dictionary.get("dataTypeUniqueID", ""),
        )


@dataclass
class InteractionDetails:
    key: str
    uniqueID: str
    name: str
    version: Optional[str]
    interactionType: InteractionType
    path: str
    spec: InteractionSpecificationSummary
    exec: Optional[InteractionExecutionSummary]
    parameters: List[ParameterSummary]
    interactions: List[InteractionDetails]

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            key=dictionary.get("key", ""),
            uniqueID=dictionary.get("uniqueID", ""),
            name=dictionary.get("name", ""),
            version=dictionary.get("version", ""),
            interactionType=InteractionType(
                dictionary.get("interactionType", InteractionType.Atomic)
            ),
            path=dictionary.get("path", ""),
            spec=InteractionSpecificationSummary.from_dict(dictionary.get("spec")),
            exec=InteractionExecutionSummary.from_dict(dictionary.get("exec"))
            if dictionary.get("exec")
            else None,
            parameters=[
                ParameterSummary.from_dict(param) for param in dictionary.get("parameters", [])
            ],
            interactions=[
                InteractionDetails.from_dict(interaction)
                for interaction in dictionary.get("interactions", [])
            ],
        )


@dataclass
class TestCaseDetails:
    uniqueID: str
    spec: TestCaseSpecificationDetails
    exec: Optional[TestCaseExecutionDetails]
    interactions: List[InteractionDetails]
    parameters: List[ParameterSummary]

    @classmethod
    def from_dict(cls, dictionary) -> TestCaseDetails:
        return cls(
            uniqueID=dictionary.get("uniqueID"),
            spec=TestCaseSpecificationDetails.from_dict(dictionary.get("spec")),
            exec=TestCaseExecutionDetails.from_dict(dictionary.get("exec"))
            if dictionary.get("exec")
            else None,
            interactions=[
                InteractionDetails.from_dict(interaction)
                for interaction in dictionary.get("interactions", [])
            ],
            parameters=[
                ParameterSummary.from_dict(parameter)
                for parameter in dictionary.get("parameters", [])
            ],
        )


@dataclass
class TestStructureSpecification:
    key: str
    locker: Optional[UserReference]
    status: SpecificationStatus

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            key=dictionary.get("key", "-1"),
            locker=UserReference.from_dict(dictionary.get("locker", {}))
            if dictionary.get("locker")
            else None,
            status=SpecificationStatus(dictionary.get("status", SpecificationStatus.Planned)),
        )


@dataclass
class TestStructureAutomation:
    key: str
    locker: Optional[UserReference]
    status: SpecificationStatus

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            key=dictionary.get("key", "-1"),
            locker=UserReference.from_dict(dictionary.get("locker", {}))
            if dictionary.get("locker")
            else None,
            status=SpecificationStatus(dictionary.get("status", SpecificationStatus.Planned)),
        )


@dataclass
class TestStructureExecution:
    key: str
    locker: Optional[UserReference]
    status: ActivityStatus
    execStatus: ExecutionStatus
    verdict: ExecutionVerdict

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            key=dictionary.get("key", "-1"),
            locker=UserReference.from_dict(dictionary.get("locker", {}))
            if dictionary.get("locker")
            else None,
            status=ActivityStatus(dictionary.get("status", ActivityStatus.Planned)),
            execStatus=ExecutionStatus(dictionary.get("execStatus", ExecutionStatus.NotBlocked)),
            verdict=ExecutionVerdict(dictionary.get("verdict", ExecutionVerdict.Undefined)),
        )


@dataclass
class AttachedFilter:
    key: str
    name: str
    filterType: FilterType
    content: str

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            key=dictionary.get("key", "-1"),
            name=dictionary.get("name", ""),
            filterType=FilterType(dictionary.get("filterType", FilterType.TestCaseSetFilter)),
            content=dictionary.get("content", ""),
        )


@dataclass
class TestStructureTreeNodeInformation:
    key: str
    numbering: str
    parentKey: str
    name: str
    uniqueID: str
    orderPos: int
    matchesFilter: bool

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            key=dictionary.get("key", "-1"),
            numbering=dictionary.get("numbering", "-1"),
            parentKey=dictionary.get("parentKey", "-1"),
            name=dictionary.get("name", ""),
            uniqueID=dictionary.get("uniqueID", ""),
            orderPos=dictionary.get("orderPos", "-1"),
            matchesFilter=dictionary.get("matchesFilter", True),
        )


@dataclass
class TestStructureTreeNode:
    elementType: TestStructureTreeNodeType
    baseInformation: TestStructureTreeNodeInformation
    specification: Optional[TestStructureSpecification]
    automation: Optional[TestStructureAutomation]
    execution: Optional[TestStructureExecution]
    filters: List[AttachedFilter]

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            elementType=TestStructureTreeNodeType(
                dictionary.get("elementType", TestStructureTreeNodeType.TestTheme)
            ),
            baseInformation=TestStructureTreeNodeInformation.from_dict(
                dictionary.get("baseInformation", {})
            ),
            specification=TestStructureSpecification.from_dict(dictionary.get("specification"))
            if dictionary.get("specification")
            else None,
            automation=TestStructureAutomation.from_dict(dictionary.get("automation"))
            if dictionary.get("automation")
            else None,
            execution=TestStructureExecution.from_dict(dictionary.get("execution"))
            if dictionary.get("execution")
            else None,
            filters=[AttachedFilter.from_dict(filter) for filter in dictionary.get("filters", [])],
        )


@dataclass
class TestStructureTree:
    root: TestStructureTreeNode
    nodes: List[TestStructureTreeNode]

    @classmethod
    def from_dict(cls, dictionary) -> TestStructureTree:
        return cls(
            root=TestStructureTreeNode.from_dict(dictionary.get("root", {}))
            if "root" in dictionary
            else None,
            nodes=[TestStructureTreeNode.from_dict(node) for node in dictionary.get("nodes", [])],
        )



# pylint: skip-file
from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Dict, List, Union

from .model import StrEnum


@dataclass
class SubdivisionsMapping:
    libraries: Dict
    resources: Dict

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            libraries=dictionary.get("libraries", {}), resources=dictionary.get("resources", {})
        )


@dataclass
class ForcedImport:
    libraries: List[str]
    resources: List[str]
    variables: List[str]

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            libraries=dictionary.get("libraries", []),
            resources=dictionary.get("resources", []),
            variables=dictionary.get("variables", []),
        )


class LogLevel(StrEnum):
    CRITICAL = "CRITICAL"
    FATAL = CRITICAL
    ERROR = "ERROR"
    WARNING = "WARNING"
    WARN = WARNING
    INFO = "INFO"
    DEBUG = "DEBUG"
    NOTSET = "NOTSET"


@dataclass
class ConsoleLoggerConfig:
    logLevel: LogLevel
    logFormat: str

    @classmethod
    def from_dict(cls, dictionary):
        log_level = dictionary.get("logLevel", "INFO").upper()
        if log_level not in LogLevel.__members__:
            print(
                f"ValueError: {log_level} is not a valid logLevel. "
                f"Available logLevel are: {list(LogLevel.__members__)}"
            )
            log_level = LogLevel.INFO
        return cls(
            logLevel=LogLevel(log_level),
            logFormat=dictionary.get("logFormat", "%(levelname)s: %(message)s"),
        )


@dataclass
class FileLoggerConfig(ConsoleLoggerConfig):
    fileName: str

    @classmethod
    def from_dict(cls, dictionary):
        log_level = dictionary.get("logLevel", "DEBUG").upper()
        if log_level not in LogLevel.__members__:
            print(
                f"ValueError: {log_level} is not a valid logLevel. "
                f"Available logLevel are: {list(LogLevel.__members__)}"
            )
            log_level = LogLevel.DEBUG
        return cls(
            logLevel=log_level,
            logFormat=dictionary.get(
                "logFormat", "%(asctime)s - %(filename)s:%(lineno)d - %(levelname)8s - %(message)s"
            ),
            fileName=dictionary.get("fileName", "testbench2robotframework.log"),
        )


@dataclass
class LoggingConfig:
    console: ConsoleLoggerConfig
    file: FileLoggerConfig

    @classmethod
    def from_dict(cls, dictionary):
        return cls(
            console=ConsoleLoggerConfig.from_dict(dictionary.get("console", {})),
            file=FileLoggerConfig.from_dict(dictionary.get("file", {})),
        )


class ReferenceBehaviour(StrEnum):
    ATTACHMENT = "ATTACHMENT"
    REFERENCE = "REFERENCE"
    NONE = "NONE"


class AttachmentConflictBehaviour(StrEnum):
    ERROR = "ERROR"
    USE_NEW = "USE_NEW"
    USE_EXISTING = "USE_EXISTING"
    RENAME_NEW = "RENAME_NEW"


@dataclass
class Configuration:
    rfLibraryRegex: List[str]
    rfResourceRegex: List[str]
    rfLibraryRoots: List[str]
    rfResourceRoots: List[str]
    fullyQualified: bool
    subdivisionsMapping: SubdivisionsMapping
    forcedImport: ForcedImport
    generationDirectory: str
    createOutputZip: bool
    resourceDirectory: str
    logSuiteNumbering: bool
    clearGenerationDirectory: bool
    loggingConfiguration: LoggingConfig
    logCompoundInteractions: bool
    testCaseSplitPathRegEx: str
    phasePattern: str
    referenceBehaviour: ReferenceBehaviour
    attachmentConflictBehaviour: AttachmentConflictBehaviour

    @classmethod
    def from_dict(cls, dictionary) -> Configuration:
        return cls(
            rfLibraryRegex=dictionary.get(
                "rfLibraryRegex", [r"(?:.*\.)?(?P<resourceName>[^.]+?)\s*\[Robot-Library\].*"]
            ),
            rfResourceRegex=dictionary.get(
                "rfResourceRegex", [r"(?:.*\.)?(?P<resourceName>[^.]+?)\s*\[Robot-Resource\].*"]
            ),
            rfLibraryRoots=dictionary.get("rfLibraryRoots", ["RF", "RF-Library"]),
            rfResourceRoots=dictionary.get("rfResourceRoots", ["RF-Resource"]),
            fullyQualified=dictionary.get("fullyQualified", False),
            subdivisionsMapping=SubdivisionsMapping.from_dict(
                dictionary.get("subdivisionsMapping", {})
            ),
            forcedImport=ForcedImport.from_dict(dictionary.get("forcedImport", {})),
            generationDirectory=dictionary.get("generationDirectory", "{root}/Generated"),
            createOutputZip=dictionary.get("createOutputZip", False),
            logSuiteNumbering=dictionary.get("logSuiteNumbering", False),
            clearGenerationDirectory=dictionary.get("clearGenerationDirectory", True),
            loggingConfiguration=LoggingConfig.from_dict(
                dictionary.get("loggingConfiguration", {})
            ),
            logCompoundInteractions=dictionary.get("logCompoundInteractions", True),
            resourceDirectory=dictionary.get("resourceDirectory", "{root}/Resources").replace(
                '\\', '/'
            ),
            testCaseSplitPathRegEx=dictionary.get("testCaseSplitPathRegEx", ".*StopWithRestart.*"),
            phasePattern=dictionary.get("phasePattern", "{testcase} : Phase {index}/{length}"),
            referenceBehaviour=ReferenceBehaviour(
                dictionary.get("referenceBehaviour", "ATTACHMENT").upper()
            ),
            attachmentConflictBehaviour=AttachmentConflictBehaviour(
                dictionary.get("attachmentConflictBehaviour", "USE_EXISTING").upper()
            ),
        )


def write_default_config(config_file):
    with open(config_file, 'w', encoding='utf-8') as file:
        json.dump(
            Configuration.from_dict({}).__dict__,
            file,
            default=lambda o: o.__dict__,
            indent=2,
        )



import json
from dataclasses import asdict
from pathlib import Path
from typing import Union

from .config import Configuration
from .log import logger
from .model import TestCaseDetails, TestCaseSetDetails, TestStructureTree

TEST_STRUCTURE_TREE_FILE = "TestThemeTree"


def write_test_structure_element(
    json_dir: str,
    test_structure_element: Union[TestStructureTree, TestCaseSetDetails, TestCaseDetails],
) -> None:
    if isinstance(test_structure_element, TestStructureTree):
        filepath = Path(json_dir) / Path(TEST_STRUCTURE_TREE_FILE + ".json")
    else:
        filepath = Path(json_dir) / Path(f"{test_structure_element.uniqueID}.json")
    with open(filepath, 'w+', encoding="utf8") as output_file:
        json.dump(asdict(test_structure_element), output_file, indent=2)


def write_default_config(config_file):
    with open(config_file, 'w+', encoding='utf-8') as file:
        json.dump(
            Configuration.from_dict({}).__dict__,
            file,
            default=lambda o: o.__dict__,
            indent=2,
        )
        logger.warning(f"Default Configuration generated to {config_file}")



import logging
from logging.handlers import RotatingFileHandler

from .config import Configuration

logger = logging.Logger("iTB2RF", logging.DEBUG)


def setup_logger(config: Configuration):
    console_handler = logging.StreamHandler()
    console_handler.setLevel(config.loggingConfiguration.console.logLevel)
    console_handler.setFormatter(logging.Formatter(config.loggingConfiguration.console.logFormat))
    logger.addHandler(console_handler)

    file_handler = RotatingFileHandler(
        filename=config.loggingConfiguration.file.fileName,
        mode="a",
        maxBytes=1 * 1024 * 1024,
        backupCount=2,
        encoding="utf_8",
        delay=False,
    )
    file_handler.setLevel(config.loggingConfiguration.file.logLevel)
    file_handler.setFormatter(logging.Formatter(config.loggingConfiguration.file.logFormat))
    logger.addHandler(file_handler)



import os
import shutil
from pathlib import Path

import pytest
from robot.parsing.model.blocks import File

from testbench2robotframework.testsuite_write import (
    clear_generation_directory,
    directory_to_zip,
    write_test_suite_files,
)


def test_previous_zip_gets_deleted():
    generation_dir = Path("./tests/test_data/Generated")
    if not os.path.isdir("./tests/test_data/Generated"):
        os.mkdir("./tests/test_data/Generated")
    shutil.make_archive(str(generation_dir), 'zip', str(generation_dir))
    clear_generation_directory(generation_dir)
    assert not os.path.exists(generation_dir)


def test_zip_gets_created():
    test_suites = {"iTB-TC-322-PC-12121212212": File()}
    generation_dir = Path("./tests/test_data/Generated/zip_file")
    write_test_suite_files(test_suites, generation_dir)
    directory_to_zip(generation_dir)
    assert os.path.exists("".join([str(generation_dir), ".zip"]))



from typing import Dict

from robot.parsing.model.blocks import File
from robot.running.model import TestSuite

from .model import TestStructureTreeNode
from .utils import PathResolver


class RobotSuiteRunner:
    def __init__(self, test_suite_models: Dict[str, File], path_resolver: PathResolver) -> None:
        self.test_suite_models = test_suite_models
        self.path_resolver = path_resolver
        self.test_suites: Dict[str, TestSuite] = {"[Root]": TestSuite()}
        for _, tcs in self.path_resolver.tcs_catalog.items():
            self._add_test_structure_element_to_test_suite_hierarchy(tcs)

    def _add_test_structure_element_to_test_suite_hierarchy(self, tse: TestStructureTreeNode):
        suite_uid = tse.baseInformation.uniqueID
        test_suite = self._get_test_suite_from_model(suite_uid, self.test_suite_models[suite_uid])
        parent_tse = self.path_resolver.tree_dict[tse.baseInformation.parentKey]
        if parent_tse.baseInformation.name != "[Root]":
            parent_uid = parent_tse.baseInformation.uniqueID
            parent_testsuite = self._get_test_suite_from_model(
                parent_uid, self.test_suite_models[parent_uid]
            )
            parent_testsuite.suites.append(test_suite)
            self._add_test_structure_element_to_test_suite_hierarchy(parent_tse)
        else:
            self.test_suites["[Root]"].suites.append(test_suite)

    def _get_test_suite_from_model(self, suite_uid: str, model: File):
        if self.test_suites.get(suite_uid):
            return self.test_suites[suite_uid]
        test_suite = TestSuite.from_model(model, name=suite_uid)
        self.test_suites[suite_uid] = test_suite
        return test_suite

    def run_suites(self):
        self.test_suites["[Root]"].run(output='./results/output.xml')



from __future__ import annotations

import os
import re
from dataclasses import dataclass
from pathlib import Path, PurePath
from typing import Dict, List, Optional, Set, Tuple
from uuid import uuid4

from robot.parsing.lexer.tokens import Token
from robot.parsing.model.blocks import File, SettingSection, TestCase, TestCaseSection
from robot.parsing.model.statements import (
    Comment,
    EmptyLine,
    ForceTags,
    KeywordCall,
    LibraryImport,
    Metadata,
    ResourceImport,
    SectionHeader,
    Statement,
    Tags,
    TestCaseName,
    VariablesImport,
)

from .config import Configuration
from .json_reader import TestCaseSet
from .log import logger
from .model import (
    InteractionDetails,
    InteractionType,
    ParameterUseType,
    TestCaseDetails,
    TestStructureTreeNode,
    UdfType,
    UserDefinedField,
)
from .utils import PathResolver, get_list_item

SEPARATOR = "    "
ROBOT_PATH_SEPARATOR = "/"
RELATIVE_RESOURCE_INDICATOR = r"^{root}"
SECTION_SEPARATOR = [EmptyLine.from_params()] * 2
TESTCASE_SEPARATOR = [EmptyLine.from_params()]
UNKNOWN_IMPORT_TYPE = str(uuid4())
LIBRARY_IMPORT_TYPE = str(uuid4())
RESOURCE_IMPORT_TYPE = str(uuid4())


@dataclass
class InteractionCall:
    name: str


@dataclass
class AtomicInteractionCall(InteractionCall):
    cbv_parameters: Dict[str, str]
    cbr_parameters: Dict[str, str]
    indent: int
    import_prefix: str


@dataclass
class CompoundInteractionCall(InteractionCall):
    cbv_parameters: Dict[str, str]
    cbr_parameters: Dict[str, str]
    indent: int


class RfTestCase:
    def __init__(self, test_case_details: TestCaseDetails, config: Configuration) -> None:
        self.uid: str = test_case_details.uniqueID
        self.interaction_calls: List[InteractionCall] = []
        self.used_imports: Dict[str, Set[str]] = {}
        self.config = config
        self.lib_pattern_list = [re.compile(pattern) for pattern in config.rfLibraryRegex]
        self.res_pattern_list = [re.compile(pattern) for pattern in config.rfResourceRegex]
        for interaction in test_case_details.interactions:
            self._get_interaction_calls(interaction)
        self.rf_tags = self._get_tags(test_case_details)
        # TODO description

    @staticmethod
    def _get_tags(test_case_details: TestCaseDetails) -> List[str]:
        tags = [keyword.name for keyword in test_case_details.spec.keywords]
        tags.extend([udf.robot_tag for udf in test_case_details.spec.udfs if udf.robot_tag])
        return tags

    @staticmethod
    def _get_udf_tags(user_defined_fields: List[UserDefinedField]) -> List[str]:
        udfs = []
        for udf in user_defined_fields:
            if udf.valueType == UdfType.Enumeration:
                udfs.append(f"{udf.name}:{udf.value}")
            elif udf.valueType == UdfType.String and udf.value:
                udfs.append(f"{udf.name}:{udf.value}")
            elif udf.valueType == UdfType.Boolean and udf.value == "true":
                udfs.append(udf.name)
        return udfs

    def _get_interaction_calls(self, interaction: InteractionDetails, indent: int = 0) -> None:
        indent += 1
        if interaction.interactionType != InteractionType.Textuell:
            cbv_params = self._get_params_by_use_type(interaction, ParameterUseType.CallByValue)
            cbr_params = self._get_params_by_use_type(
                interaction,
                ParameterUseType.CallByReference,
                ParameterUseType.CallByReferenceMandatory,
            )
            if interaction.interactionType == InteractionType.Compound:
                self._append_compound_ia_and_analyze_children(
                    cbr_params, cbv_params, indent, interaction
                )
            elif interaction.interactionType == InteractionType.Atomic:
                self._append_atomic_ia(
                    cbr_params, cbv_params, indent, interaction
                )  # TODO: Else für textuelle Interaktionen

    def _append_atomic_ia(
        self,
        cbr_params: Dict[str, str],
        cbv_params: Dict[str, str],
        indent: int,
        interaction: InteractionDetails,
    ):
        resource_type, import_prefix = self._get_keyword_import(interaction)

        if resource_type not in self.used_imports:
            self.used_imports[resource_type] = {import_prefix}
        else:
            self.used_imports[resource_type].add(import_prefix)
        self.interaction_calls.append(
            AtomicInteractionCall(
                name=interaction.name,
                cbv_parameters=cbv_params,
                cbr_parameters=cbr_params,
                indent=indent,
                import_prefix=import_prefix,
            )
        )

    def _get_keyword_import(self, interaction) -> Tuple[str, str]:
        for pattern in self.lib_pattern_list:
            match = pattern.search(interaction.path)
            if match:
                return LIBRARY_IMPORT_TYPE, match[1].strip()
        for pattern in self.res_pattern_list:
            match = pattern.search(interaction.path)
            if match:
                return RESOURCE_IMPORT_TYPE, match[1].strip()

        ia_path_parts = interaction.path.split(".")
        if len(ia_path_parts) == 1:
            return UNKNOWN_IMPORT_TYPE, ia_path_parts[0]
        root_subdivision, import_prefix = ia_path_parts[:2]
        if root_subdivision in self.config.rfLibraryRoots:
            return LIBRARY_IMPORT_TYPE, import_prefix
        if root_subdivision in self.config.rfResourceRoots:
            return RESOURCE_IMPORT_TYPE, import_prefix

        return root_subdivision, import_prefix

    def _append_compound_ia_and_analyze_children(
        self,
        cbr_params: Dict[str, str],
        cbv_params: Dict[str, str],
        indent: int,
        interaction_detail: InteractionDetails,
    ):
        self.interaction_calls.append(
            CompoundInteractionCall(
                interaction_detail.name,
                cbv_parameters=cbv_params,
                cbr_parameters=cbr_params,
                indent=indent,
            )
        )
        for interaction in interaction_detail.interactions:
            self._get_interaction_calls(interaction, indent)

    def _create_rf_keywords(
        self, interaction_calls: List[InteractionCall]
    ) -> List[List[Statement]]:
        keyword_lists: List[List[Statement]] = [[]]
        tc_index = 0
        is_first_atomic = True
        for interaction_call in interaction_calls:
            if isinstance(interaction_call, AtomicInteractionCall):
                if self.is_splitting_ia(interaction_call, keyword_lists, tc_index):
                    if not is_first_atomic:
                        tc_index += 1
                        keyword_lists.append([])
                is_first_atomic = False
                keyword_lists[tc_index].append(self._create_rf_keyword(interaction_call))
            elif (
                isinstance(interaction_call, CompoundInteractionCall)
                and self.config.logCompoundInteractions
            ):
                keyword_lists[tc_index].append(self._create_rf_compound_keyword(interaction_call))
        return keyword_lists

    def is_splitting_ia(self, interaction_call, keyword_lists, tc_index):
        return (
            re.search(
                self.config.testCaseSplitPathRegEx,
                f"{interaction_call.import_prefix}.{interaction_call.name}",
            )
            and keyword_lists[tc_index]
        )

    def to_robot_ast_test_cases(
        self,
    ) -> List[
        TestCase
    ]:  # TODO: Separate testcase splitting from this method --> new method: to_robot_ast_test_case
        rf_keyword_lists = self._create_rf_keywords(self.interaction_calls)
        rf_test_cases: List[TestCase] = []
        multiple_tests = len(rf_keyword_lists) > 1
        for index, rf_keywords in enumerate(rf_keyword_lists):
            phase_pattern = self.config.phasePattern
            tc_name = (
                phase_pattern.format(
                    testcase=self.uid, index=index + 1, length=len(rf_keyword_lists)
                )
                if multiple_tests
                else self.uid
            )
            # suffix = f' : Phase {index + 1}/{len(rf_keyword_lists)}' if multiple_tests else ''
            # tc_name = f"{self.uid}{suffix}"  # TODO later UID or Comments
            rf_test_case = TestCase(header=TestCaseName.from_params(tc_name))
            rf_test_case.body.append(Tags.from_params(self.rf_tags))
            rf_test_case.body.extend(rf_keywords)
            rf_test_case.body.extend(TESTCASE_SEPARATOR)
            rf_test_cases.append(rf_test_case)
        return rf_test_cases

    @staticmethod
    def _create_cbv_parameters(interaction: AtomicInteractionCall) -> List[str]:
        parameters = []
        for name, value in interaction.cbv_parameters.items():
            if name.endswith('='):
                parameters.append(f"{name}{value}")
            else:
                parameters.append(value)
        return parameters

    @staticmethod
    def _create_cbr_parameters(interaction: AtomicInteractionCall) -> List[str]:
        cbr_parameters = list(interaction.cbr_parameters.values())
        for index, parameter in enumerate(cbr_parameters):
            if not parameter.startswith('${'):
                cbr_parameters[index] = f"${{{parameter}}}"
        return cbr_parameters

    def _create_rf_keyword(self, interaction: AtomicInteractionCall) -> KeywordCall:
        import_prefix = (self.config.fullyQualified or False) * f"{interaction.import_prefix}."
        interaction_indent = (
            SEPARATOR * interaction.indent if self.config.logCompoundInteractions else SEPARATOR
        )
        cbv_parameters = self._create_cbv_parameters(interaction)
        cbr_parameters = self._create_cbr_parameters(interaction)
        return KeywordCall.from_params(
            assign=tuple(cbr_parameters),
            name=f"{import_prefix}{interaction.name}",
            args=tuple(cbv_parameters),
            indent=interaction_indent,
        )

    def _create_rf_compound_keyword(self, interaction: CompoundInteractionCall) -> Comment:
        interaction_indent = " " * (interaction.indent * 4)
        return Comment.from_params(
            comment=self._generate_compound_interaction_comment(interaction),
            indent=interaction_indent,
        )  # TODO  prio later key=value als named erlauben config?

    @staticmethod
    def _generate_compound_interaction_comment(interaction: CompoundInteractionCall) -> str:
        cbr_params = SEPARATOR.join(
            [
                f"{param_name}={param_value}"
                for param_name, param_value in interaction.cbr_parameters.items()
            ]
        )
        cbv_params = SEPARATOR.join(
            [
                f"{param_name}={param_value}"
                for param_name, param_value in interaction.cbv_parameters.items()
            ]
        )
        if cbr_params:
            return f"#{SEPARATOR.join([cbr_params, interaction.name, cbv_params])}"
        return f"#{SEPARATOR.join([interaction.name, cbv_params])}"

    @staticmethod
    def _get_params_by_use_type(
        interaction: InteractionDetails, *param_use_types: ParameterUseType
    ) -> Dict[str, str]:
        return {
            parameter.name: parameter.value
            for parameter in interaction.parameters
            if parameter.parameterUseType in param_use_types
        }


def create_test_suites(
    test_case_set_catalog: Dict[str, TestCaseSet],
    path_resolver: PathResolver,
    config: Configuration,
) -> Dict[str, File]:
    tcs_paths = path_resolver.tcs_paths
    test_suites = {}
    for uid, test_case_set in test_case_set_catalog.items():
        test_suites[uid] = RobotSuiteFileBuilder(
            test_case_set, tcs_paths[uid], config
        ).create_test_suite_file()
    tt_paths = path_resolver.tt_paths
    for uid, test_theme in path_resolver.tt_catalog.items():
        test_suites[uid] = RobotInitFileBuilder(
            test_theme, tt_paths[uid], config
        ).create_init_file()
    return test_suites


class RobotInitFileBuilder:
    def __init__(
        self, test_theme: TestStructureTreeNode, tt_path: PurePath, config: Configuration
    ) -> None:
        self.test_theme = test_theme
        self.tt_path = tt_path
        self.config = config

    def create_init_file(self) -> File:
        sections = [self._create_setting_section()]
        return File(sections, source=os.path.join(str(self.tt_path), "__init__"))

    def _create_setting_section(self) -> SettingSection:
        setting_section = SettingSection(header=SectionHeader.from_params(Token.SETTING_HEADER))
        setting_section_meta_data = self._get_setting_section_metadata()
        setting_section.body.extend(
            [
                create_meta_data(metadata_name, metadata_value)
                for metadata_name, metadata_value in setting_section_meta_data.items()
            ]
        )
        return setting_section

    def _get_setting_section_metadata(self) -> Dict[str, str]:
        meta_data = {
            "UniqueID": self.test_theme.baseInformation.uniqueID,
            "Numbering": self.test_theme.baseInformation.numbering,
        }
        if self.test_theme.specification:
            meta_data["Specification Status"] = self.test_theme.specification.status
        return meta_data


def create_meta_data(name, value):
    tokens = [
        Token(Metadata, 'Metadata', 1),
        Token(SEPARATOR, '    ', 2),
        Token('NAME', name, 3),
        Token(SEPARATOR, '    ', 4),
        Token('ARGUMENT', value, 5),
        Token('EOL', '\n', 6),
    ]
    return Metadata(tokens)


class RobotSuiteFileBuilder:
    def __init__(
        self, test_case_set: TestCaseSet, tcs_path: PurePath, config: Configuration
    ) -> None:
        self.test_case_set = test_case_set
        self.tcs_path = tcs_path
        self.config = config
        self._rf_test_cases: List[RfTestCase] = [
            RfTestCase(test_case_details=test_case, config=config)
            for test_case in self.test_case_set.test_cases.values()
        ]

    def create_test_suite_file(self) -> File:
        sections = [self._create_setting_section(), self._create_test_case_section()]
        return File(sections, source=str(self.tcs_path))

    def _create_test_case_section(self) -> TestCaseSection:
        test_case_section = TestCaseSection(header=SectionHeader.from_params(Token.TESTCASE_HEADER))
        robot_ast_test_cases = []
        for test_case in self._rf_test_cases:
            robot_ast_test_cases.extend(test_case.to_robot_ast_test_cases())
        test_case_section.body.extend(robot_ast_test_cases)
        test_case_section.body.extend(SECTION_SEPARATOR)
        return test_case_section

    def _get_used_subdivisions(self) -> Dict[str, Set[str]]:
        import_dict: Dict[str, Set[str]] = {}
        for test_case in self._rf_test_cases:
            for root, import_names in test_case.used_imports.items():
                if root not in import_dict:
                    import_dict[root] = import_names
                else:
                    import_dict[root].update(import_names)
        return import_dict

    def _create_rf_variable_imports(self) -> List[VariablesImport]:
        return [
            VariablesImport.from_params(name=variable_file)
            for variable_file in self.config.forcedImport.variables
        ]

    def _create_rf_resource_imports(self, import_dict: Dict[str, Set[str]]) -> List[ResourceImport]:
        resources = {
            str(resource)
            for resource_root, resources in import_dict.items()
            for resource in resources
            if not self._is_library(resource_root) and self._is_resource(resource_root)
        }
        resources.update(self.config.forcedImport.resources)
        resource_paths = {
            self._create_resource_path(resource) for resource in sorted(resources)
        }  # TODO Fix Paths to correct models
        return [ResourceImport.from_params(res) for res in sorted(resource_paths)]

    def _create_resource_path(self, resource: str) -> str:
        if (
            not self.config.resourceDirectory
        ):  # TODO Fehler: Hier springt er raus weil resourceDir nicht gesetzt, sollte aber mindestens das Mapping noch mitnehmen.
            return f"{resource}.resource"
        if not re.match(RELATIVE_RESOURCE_INDICATOR, self.config.resourceDirectory):
            return f"{self.config.resourceDirectory}{ROBOT_PATH_SEPARATOR}{resource}.resource"
        resource_dir = self._get_resource_directory()
        resource_path = self.config.subdivisionsMapping.resources.get(resource)
        resource_base_path = (
            f"{('..' + ROBOT_PATH_SEPARATOR) * len(self.tcs_path.parts)}{resource_dir}"
        )
        if resource_path:
            return f"{resource_base_path}{ROBOT_PATH_SEPARATOR}{resource_path}"
        return f"{resource_base_path}{ROBOT_PATH_SEPARATOR}{resource}.resource"

    def _get_resource_directory(self) -> str:
        root_path = Path(os.curdir).absolute()
        return os.path.relpath(
            re.sub(
                RELATIVE_RESOURCE_INDICATOR,
                str(root_path).replace('\\', '\\\\'),
                self.config.resourceDirectory,
                flags=re.IGNORECASE,
            ),
            root_path,
        ).replace('\\', ROBOT_PATH_SEPARATOR)

    @staticmethod
    def _is_library(root_subdivision: str) -> bool:
        return root_subdivision == LIBRARY_IMPORT_TYPE

    @staticmethod
    def _is_resource(root_subdivision: str) -> bool:
        return root_subdivision == RESOURCE_IMPORT_TYPE

    def _create_rf_library_imports(self, import_dict: Dict[str, Set[str]]) -> List[LibraryImport]:
        libraries = {
            str(library)
            for library_root, libraries in import_dict.items()
            for library in libraries
            if self._is_library(library_root)
        }
        libraries.update(self.config.forcedImport.libraries)
        lib_imports = {
            self.config.subdivisionsMapping.libraries.get(library, library) for library in libraries
        }
        return [LibraryImport.from_params(lib) for lib in sorted(lib_imports)]

    def _create_rf_force_tags(self) -> Optional[ForceTags]:
        tb_keyword_names = [keyword.name for keyword in self.test_case_set.details.spec.keywords]
        udfs = [udf.robot_tag for udf in self.test_case_set.details.spec.udfs if udf.robot_tag]
        force_tags = tb_keyword_names + udfs
        if force_tags:
            return ForceTags.from_params(force_tags)
        return None

    def _create_rf_unknown_imports(self, import_dict: Dict[str, Set[str]]) -> List[Comment]:
        unknown_imports = {
            str(unknown_import)
            for root_subdivision, unknown_imports in import_dict.items()
            for unknown_import in unknown_imports
            if not self._is_library(root_subdivision) and not self._is_resource(root_subdivision)
        }
        for root, subdivision_names in import_dict.items():
            logger.debug(
                f"{self.test_case_set.details.uniqueID} has imports {list(subdivision_names)} "
                f"from unknown root subdivision '{root}'!"
            )
        if unknown_imports:
            logger.warning(
                f"{self.test_case_set.details.uniqueID} has unknown imports. "
                f"TestBench Subdivisions which correspond to Libraries or Resources "
                f"need to be declared in the configuration file. "
                f"See Log for more details."
            )
        return [
            Comment.from_params(comment=f"# UNKNOWN    {unknown}", indent="")
            for unknown in unknown_imports
        ]

    def _create_setting_section(self) -> SettingSection:
        subdivisions = self._get_used_subdivisions()
        setting_section = SettingSection(header=SectionHeader.from_params(Token.SETTING_HEADER))
        setting_section.body.extend(self._create_rf_variable_imports())
        setting_section.body.extend(self._create_rf_library_imports(subdivisions))
        setting_section.body.extend(self._create_rf_resource_imports(subdivisions))
        setting_section.body.extend(self._create_rf_unknown_imports(subdivisions))
        setting_section_meta_data = self.test_case_set.metadata
        setting_section.body.extend(
            [
                create_meta_data(metadata_name, metadata_value)
                for metadata_name, metadata_value in setting_section_meta_data.items()
            ]
        )
        setting_section.body.append(self._create_rf_force_tags())
        setting_section.body.extend(SECTION_SEPARATOR)
        return setting_section



import argparse
import os
import re
import shutil
import sys
from pathlib import Path, PurePath
from typing import Dict, Optional, Tuple
from zipfile import ZipFile

from testbench2robotframework.model import (
    TestStructureTree,
    TestStructureTreeNode,
    TestStructureTreeNodeType,
)

CONVERTER_DESCRIPTION = """tB2Robot converts TestBench JSON report to Robot Framework Code
                        and Robot Result Model to JSON full report."""
WRITE_SUBPARSER_HELP = """Command to convert TestBench´s JSON REPORT to Robot Framework Code."""
READ_SUBPARSER_HELP = """Command to read a robot output xml file and
write the results to a TestBench JSON REPORT."""
JSON_PATH_ARGUMENT_HELP = "Path to a ZIP file or directory containing TestBenchs JSON report files."
CONFIG_ARGUMENT_HELP = """Path to a config json file to generate robot files
                        based on the given configuration.
                        If no path is given testbench2robot will search for a file
                        named \"config.json\" in the current working directory."""
ROBOT_OUTPUT_HELP = """Path to an XML file containing the robot results."""
ROBOT_RESULT_HELP = """Path to the directory or ZIP File the TestBench JSON reports
with result should be saved to."""


arg_parser = argparse.ArgumentParser(description=CONVERTER_DESCRIPTION)
arg_parser.add_argument(
    '--version',
    '--info',
    action='store_true',
    help='Writes the TestBench2RobotFramework, Robot Framework and Python version to console.',
)
subparsers = arg_parser.add_subparsers(dest="subcommand")

write_parser = subparsers.add_parser("write", help=WRITE_SUBPARSER_HELP)
write_parser.add_argument(
    "-c",
    "--config",
    help=CONFIG_ARGUMENT_HELP,
    type=str,
    required=False,
    default=os.path.join(os.path.abspath(os.curdir), "config.json"),
)

write_parser.add_argument("jsonReport", nargs=1, type=str, help=JSON_PATH_ARGUMENT_HELP)

read_parser = subparsers.add_parser("read", help=READ_SUBPARSER_HELP)
read_parser.add_argument(
    "-c",
    "--config",
    help=CONFIG_ARGUMENT_HELP,
    type=str,
    required=False,
    default=os.path.join(os.path.abspath(os.curdir), "config.json"),
)
read_parser.add_argument(
    "-r",
    "--result",
    help=ROBOT_RESULT_HELP,
    type=str,
    required=False,
)
required_named_arguments = read_parser.add_argument_group('required named arguments')
required_named_arguments.add_argument(
    "-o", "--output", help=ROBOT_OUTPUT_HELP, type=str, required=True
)

read_parser.add_argument("jsonReport", nargs=1, type=str, help=JSON_PATH_ARGUMENT_HELP)


class PathResolver:
    def __init__(
        self,
        test_theme_tree: TestStructureTree,
        uids_of_existing_tcs: Tuple[str, ...],
        log_suite_numbers: bool,
    ):
        self.tcs_catalog: Dict[str, TestStructureTreeNode] = {}
        self.tt_catalog: Dict[str, TestStructureTreeNode] = {}
        self.tree_dict: Dict[str, TestStructureTreeNode] = {}
        self._last_child_indices: Dict[str, int] = {}
        self._log_suite_numbers = log_suite_numbers
        self._uids_of_existing_tcs = uids_of_existing_tcs
        self._analyze_tree(test_theme_tree)
        self.tcs_paths = self._get_paths(self.tcs_catalog)
        self.tt_paths = self._get_paths(self.tt_catalog)

    def _analyze_tree(self, test_theme_tree: TestStructureTree):
        self.tree_dict[test_theme_tree.root.baseInformation.key] = test_theme_tree.root
        for tse in test_theme_tree.nodes:
            self._add_existing_tcs_to_catalog(tse)
            self.tree_dict[tse.baseInformation.key] = tse
            self._store_highest_child_index(tse)

    def _store_highest_child_index(self, tse):
        self._last_child_indices[tse.baseInformation.parentKey] = max(
            int(get_tse_index(tse)),
            self._last_child_indices.get(tse.baseInformation.parentKey, 0),
        )

    def _add_existing_tcs_to_catalog(self, tse):
        if (
            tse.elementType == TestStructureTreeNodeType.TestCaseSet
            and tse.baseInformation.uniqueID in self._uids_of_existing_tcs
        ):
            self.tcs_catalog[tse.baseInformation.uniqueID] = tse

    def _get_paths(self, tse_catalog: Dict[str, TestStructureTreeNode]) -> Dict[str, PurePath]:
        return {uid: self._resolve_tse_path(tse) for uid, tse in tse_catalog.items()}

    def _resolve_tse_path(self, tse: TestStructureTreeNode) -> PurePath:
        self._add_tt_to_tt_catalog(tse)
        if tse.elementType == TestStructureTreeNodeType.Root:
            return PurePath()
        tse_name = replace_invalid_characters(tse.baseInformation.name)
        if tse.baseInformation.parentKey not in self.tree_dict:
            return PurePath(f"{self._file_prefix(tse)}{tse_name}")
        parent_path = self._resolve_tse_path(self.tree_dict[tse.baseInformation.parentKey])
        return parent_path / f"{self._file_prefix(tse)}{tse_name}"

    def _add_tt_to_tt_catalog(self, tse):
        if (
            tse.elementType == TestStructureTreeNodeType.TestTheme
            and tse.baseInformation.uniqueID not in self.tt_catalog
        ):
            self.tt_catalog[tse.baseInformation.uniqueID] = tse

    def _file_prefix(self, tse) -> str:
        prefix_separator = '_' * self._log_suite_numbers
        return f"{self._get_padded_index(tse)}_{prefix_separator}"

    def _get_padded_index(self, tse) -> str:
        index = get_tse_index(tse)
        max_length = len(str(self._last_child_indices[tse.baseInformation.parentKey]))
        return index.zfill(max_length)


def get_directory(json_report_path: Optional[str]) -> str:
    if json_report_path is None:
        return ""
    if not os.path.exists(json_report_path):
        sys.exit("Error opening " + json_report_path + ". Path does not exist.")
    if os.path.isdir(json_report_path):
        return os.path.abspath(json_report_path)
    filename, ext = os.path.splitext(json_report_path)
    if ext.lower() == ".zip":
        with ZipFile(json_report_path, 'r') as zip_ref:
            zip_ref.extractall(filename)
        return os.path.abspath(filename)
    sys.exit("Error opening " + json_report_path + ". File is not a ZIP file.")


def ensure_dir_exists(cli_output_dir):
    if not Path(cli_output_dir).is_dir():
        Path(cli_output_dir).mkdir(parents=True, exist_ok=True)


def replace_invalid_characters(name: str) -> str:
    return re.sub(r'[<>:"/\\|?* ]', "_", name)


def get_tse_index(tse: TestStructureTreeNode) -> str:
    return tse.baseInformation.numbering.rsplit(".", 1)[-1]


def directory_to_zip(directory: Path, new_path: str = None):
    if new_path:
        shutil.make_archive(str(new_path), 'zip', str(directory))
    else:
        shutil.make_archive(str(directory), 'zip', str(directory))


def get_list_item(liste, index, default: Optional[str]):
    try:
        item = liste[index]
        return item
    except IndexError:
        return default



import pytest

from testbench2robotframework.utils import get_directory


def test_json_dir_does_not_exist():
    json_files_path = "invalid/file/path"
    with pytest.raises(SystemExit):
        get_directory(json_files_path)



import html
import os
import re
import shutil
import tempfile
import uuid
from datetime import datetime, timedelta
from pathlib import Path
from shutil import copytree
from typing import Dict, List, Optional, Union
from urllib.parse import unquote

from robot.result import Keyword, ResultVisitor, TestCase, TestSuite

from .config import AttachmentConflictBehaviour, Configuration, ReferenceBehaviour
from .json_reader import TestBenchJsonReader
from .json_writer import write_test_structure_element
from .log import logger
from .model import (
    ActivityStatus,
    ExecutionVerdict,
    InteractionDetails,
    InteractionExecutionSummary,
    InteractionType,
    InteractionVerdict,
    Reference,
    ReferenceType,
    TestCaseDetails,
    TestCaseExecutionDetails,
)
from .utils import directory_to_zip, ensure_dir_exists, get_directory

BACKGROUND_COLOR = {
    "PASS": "#04AF91",
    "FAIL": "#ce3e01",
    "ERROR": "#ce3e01",
    "SKIP": "#F3E96A",
    "WARN": "#D48627",
    "INFO": "#ddd",
}
COLOR = {
    "PASS": "#fff",
    "FAIL": "#fff",
    "ERROR": "#fff",
    "SKIP": "#000",
    "WARN": "#fff",
    "INFO": "#000",
}


class ResultWriter(ResultVisitor):
    def __init__(
        self, json_report: str, json_result: str, config: Configuration, output_xml
    ) -> None:
        self.json_dir = get_directory(json_report)
        self.output_xml = output_xml
        self.reference_behaviour = config.referenceBehaviour
        self.attachment_conflict_behaviour = config.attachmentConflictBehaviour
        self.tempdir = tempfile.TemporaryDirectory(dir=os.curdir)
        if json_result is None:
            self.create_zip = bool(os.path.splitext(json_report)[1].lower() == ".zip")
            self.json_result = self.json_dir
            self.json_result_path = self.json_dir
        else:
            self.json_result_path, result_extension = os.path.splitext(json_result)
            self.create_zip = bool(result_extension.lower() == ".zip")
            self.json_result = self.tempdir.name
            if self.create_zip:
                copytree(self.json_dir, self.json_result, dirs_exist_ok=True)
        self.json_reader = TestBenchJsonReader(self.json_dir)
        self.attachments_path = Path(self.json_result, "attachments")
        # if self.attachments_path.exists():  TODO: RR Sollten wir löschen????
        #     shutil.rmtree(self.attachments_path)
        # os.mkdir(self.attachments_path)
        self.test_suites: Dict[str, TestSuite] = {}
        self.keywords: List[Keyword] = []
        self.itb_test_case_catalog: Dict[str, TestCaseDetails] = {}
        self.phase_pattern = config.phasePattern
        self.test_chain: List[TestCase] = []

    def start_suite(self, suite: TestSuite):
        if suite.metadata:
            self.test_suites[suite.metadata["uniqueID"]] = suite

    def _get_interactions_by_type(
        self, interactions: List[InteractionDetails], interaction_type: InteractionType
    ):
        for interaction in interactions:
            if interaction.interactionType == interaction_type:
                yield interaction
            if interaction.interactionType == InteractionType.Compound:
                yield from self._get_interactions_by_type(
                    interaction.interactions, interaction_type
                )

    def end_test(self, test: TestCase):
        test_chain = get_test_chain(test.name, self.phase_pattern)
        if test_chain:
            if test_chain.index == 1:
                self.test_chain = [test]
            else:
                self.test_chain.append(test)
            if test_chain.index != test_chain.length:
                return
        else:
            self.test_chain = [test]

        test_uid = test_chain.name if test_chain else test.name
        itb_test_case = self.json_reader.read_test_case(test_uid)  # TODO What if name != UID
        if itb_test_case.exec is None:
            itb_test_case.exec = TestCaseExecutionDetails.from_dict({})
        if itb_test_case.exec.key in ["", "-1"]:
            logger.warning(
                f"Test case {itb_test_case.uniqueID} was not exported based on "
                f"execution and is therefore not importable."
            )
        try:
            atomic_interactions = list(
                self._get_interactions_by_type(itb_test_case.interactions, InteractionType.Atomic)
            )
            compound_interactions = list(
                self._get_interactions_by_type(itb_test_case.interactions, InteractionType.Compound)
            )
            self._set_atomic_interactions_execution_result(atomic_interactions, self.test_chain)
            for interaction in compound_interactions:
                self._set_compound_interaction_execution_result(interaction)
            self._set_itb_testcase_execution_result(itb_test_case, self.test_chain)
            self._set_itb_testcase_execution_comment(itb_test_case, self.test_chain)
            if self.reference_behaviour != ReferenceBehaviour.NONE:
                self._set_itb_testcase_references(itb_test_case, self.test_chain)
        except TypeError:
            logger.error(
                "Could not find an itb testcase that corresponds "
                "to the given Robot Framework testcase."
            )
        self.itb_test_case_catalog[test_uid] = itb_test_case
        write_test_structure_element(self.json_result, itb_test_case)
        logger.debug(
            f"Successfully wrote the result from test "
            f"{itb_test_case.uniqueID} to TestBench's Json Report."
        )

    def _set_itb_testcase_references(
        self, itb_test_case: TestCaseDetails, test_chain: List[TestCase]
    ):
        for test in test_chain:
            itb_references = self._get_itb_reference(test.message)
            for reference in itb_references:
                if reference not in itb_test_case.exec.references:
                    itb_test_case.exec.references.append(reference)

    def _get_itb_reference(self, test_message: str) -> List[Reference]:
        references = []
        for path in re.findall(r"itb-reference:\s*(\S*)", test_message):
            if path.startswith("file:///"):
                file_path = Path(unquote(path[len("file:///") :]))
                output_dir = Path(self.output_xml).parent
                if file_path.exists():
                    reference_path = file_path
                elif Path(output_dir, file_path).exists():
                    reference_path = Path(output_dir, file_path)
                else:
                    if (
                        file_path.is_absolute()
                        and self.reference_behaviour == ReferenceBehaviour.REFERENCE
                    ):
                        references.append(self._create_reference(file_path))
                    else:
                        logger.warning(f"Referenced file '{file_path}' does not exist.")
                    continue
                file_size = os.path.getsize(reference_path)
                if file_size >= 10000000:
                    logger.error(
                        f"Trying to attach file '{reference_path}'. "
                        "Attachment file size must not exceed 10 MB "
                        f"but is {file_size / 1000000} MB."
                    )
                    reference = self._create_reference(reference_path.name)
                else:
                    reference = self._create_attachment(reference_path)
                if reference:
                    references.append(reference)
            elif re.match(r"\S+://", path):
                references.append(Reference(ReferenceType.Hyperlink, path))
            else:
                logger.warning(f"Could not identify type of test message reference '{path}'.")
        return references

    @staticmethod
    def _create_reference(reference_path: Union[Path, str]) -> Reference:
        return Reference(ReferenceType.Reference, str(reference_path))

    def _create_attachment(self, filepath: Path) -> Optional[Reference]:
        if self.reference_behaviour == ReferenceBehaviour.REFERENCE:
            return self._create_reference(filepath.resolve())
        ensure_dir_exists(self.attachments_path)
        filename = Path(filepath).name
        if (
            not Path(self.attachments_path, filename).exists()
            or self.attachment_conflict_behaviour == AttachmentConflictBehaviour.USE_NEW
        ):
            shutil.copyfile(filepath, self.attachments_path / filename, follow_symlinks=True)
            return Reference(ReferenceType.Attachment, f"attachments/{filename}")
        if self.attachment_conflict_behaviour == AttachmentConflictBehaviour.USE_EXISTING:
            return Reference(ReferenceType.Attachment, f"attachments/{filename}")
        if self.attachment_conflict_behaviour == AttachmentConflictBehaviour.RENAME_NEW:
            unique_path = self._create_unique_path(self.attachments_path / filename)
            shutil.copyfile(filepath, unique_path, follow_symlinks=True)
            unique_file = Path(unique_path).name
            logger.info(
                f"Attachment '{filename}' does already exist. "
                f"Creating new unique attachment '{unique_file}'."
            )
            return Reference(ReferenceType.Attachment, f"attachments/{unique_file}")
        if self.attachment_conflict_behaviour == AttachmentConflictBehaviour.ERROR:
            logger.error(f"Attachment '{filename}' does already exist.")
            return None

    @staticmethod
    def _create_unique_path(attachement_path: Path) -> Path:
        counter = 1
        while attachement_path.exists():
            attachement_path = Path(f"{attachement_path.stem}_{counter}{attachement_path.suffix}")
            counter += 1
        return attachement_path

    def _set_itb_testcase_execution_comment(self, itb_test_case, test_chain: List[TestCase]):
        exec_comments = []
        for test in test_chain:
            message = re.sub(r"\s*itb-reference:\s*(\S*)", "", test.message)
            html_message = (
                message[len("*HTML*") :].replace("<hr>", "<br/>").replace("<br>", "<br/>").strip()
                if test.message.startswith("*HTML*")
                else html.escape(message)
            )
            test_chain_obj = get_test_chain(test.name, self.phase_pattern)
            test_phase_name = (
                f"<b>Phase {test_chain_obj.index}/{test_chain_obj.length} : <span {self.render_status(test.status)}>{test.status}</span></b>"
                if test_chain_obj
                else f"<b><span {self.render_status(test.status)}>{test.status}</span></b>"
            )
            exec_comment = (
                f"{test_phase_name}"
                "<pre>"
                f"Start Time:   {self.get_isotime_from_robot_timestamp(test.starttime)}\n"
                f"End Time:     {self.get_isotime_from_robot_timestamp(test.endtime)}\n"
                f"Elapsed Time: {str(timedelta(milliseconds=test.elapsedtime))}\n"
                "</pre>"
                f"Message: <p><pre>{html_message}<pre></p>\n"
            )
            exec_comments.append(exec_comment)
        itb_test_case.exec.comments = f"<html><body>{''.join(exec_comments)}</body></html>"

    def _set_itb_testcase_execution_result(self, itb_test_case, test_chain):
        has_failed_chain = list(filter(lambda tc: tc.status.upper() == "FAIL", test_chain))
        passed_keywords = all(tc.status.upper() == "PASS" for tc in test_chain)
        itb_test_case.exec.actualDuration = sum([tc.elapsedtime for tc in test_chain])
        if has_failed_chain:
            self._set_itb_test_case_status(itb_test_case, "fail")
        elif passed_keywords:
            self._set_itb_test_case_status(itb_test_case, "pass")
        else:
            self._set_itb_test_case_status(itb_test_case, "undef")

    def _set_atomic_interactions_execution_result(
        self, atomic_interactions: List[InteractionDetails], test_chain: List[TestCase]
    ):
        test_chain_body = [keyword for test_phase in test_chain for keyword in test_phase.body]
        for index, interaction in enumerate(atomic_interactions):
            if interaction.exec is None:
                interaction.exec = InteractionExecutionSummary.from_dict({})
            if index < len(test_chain_body):
                keyword = test_chain_body[index]
                if not is_normalized_equal(
                    keyword.kwname, interaction.name
                ) and not is_normalized_equal(keyword.kwname.split('.')[-1], interaction.name):
                    raise NameError(
                        f"Execution can not be parsed, "
                        f"because keyword name '{keyword.kwname}' does not match with "
                        f"interaction '{interaction.name}' name."
                    )
                interaction.exec.verdict = self._get_interaction_result(keyword.status)
                interaction.exec.time = keyword.endtime
                interaction.exec.duration = keyword.elapsedtime
                interaction.exec.comments = self.get_html_keyword_comment(keyword)
            else:
                interaction.exec.verdict = InteractionVerdict.Undefined

    def get_html_keyword_comment(self, keyword: Keyword):
        messages = [self._create_itb_exec_comment(message) for message in keyword.messages]
        return (
            "<html>"
            "<body>"
            "<style>"
            "td {padding: 5px; border: none;} "
            "table {font-family: monospace; border: none;}"
            "</style>"
            "<pre>"
            f"Start Time:   {self.get_isotime_from_robot_timestamp(keyword.starttime)}\n"
            f"End Time:     {self.get_isotime_from_robot_timestamp(keyword.endtime)}\n"
            f"Elapsed Time: {str(timedelta(milliseconds=keyword.elapsedtime))}\n"
            "</pre>"
            "<table>"
            "<tr>"
            f"{'</tr><tr>'.join(messages)}"
            "</tr>"
            "</table>"
            "</body>"
            "</html>"
        )

    def _create_itb_exec_comment(
        self,
        message,
    ) -> str:  # Todo: low prio: pattern für message in config festlegen
        message_time = self.get_isotime_from_robot_timestamp(
            message.timestamp, time_format="%H:%M:%S.%f"
        )

        return (
            f"<td {self.render_status(message.level)}><b>{message.level}</b></td>"
            f"<td><pre>{message.message}</pre></td><td>{message_time}</td>"
        )

    @staticmethod
    def get_isotime_from_robot_timestamp(timestamp, time_format="%Y-%m-%d %H:%M:%S.%f"):
        return datetime.strptime(timestamp, "%Y%m%d %H:%M:%S.%f").strftime(time_format)[:-3]

    def _set_compound_interaction_execution_result(self, compound_interaction: InteractionDetails):
        atomic_interactions = list(
            self._get_interactions_by_type(
                compound_interaction.interactions, InteractionType.Atomic
            )
        )
        if compound_interaction.exec is None:
            compound_interaction.exec = InteractionExecutionSummary.from_dict({})
        compound_interaction.exec.verdict = InteractionVerdict.Skipped
        for atomic in atomic_interactions:
            if atomic.exec is None:
                logger.debug(
                    f"Atomic interaction {atomic.uniqueID} "
                    f"had no exec details and therefore ignored."
                )
                atomic.exec = InteractionExecutionSummary.from_dict({})
                # continue
            if atomic.exec.verdict is InteractionVerdict.Fail:
                compound_interaction.exec.verdict = InteractionVerdict.Fail
                break
            if atomic.exec.verdict is InteractionVerdict.Pass:
                compound_interaction.exec.verdict = InteractionVerdict.Pass

        compound_interaction.exec.duration = sum(
            [interaction.exec.duration for interaction in atomic_interactions]
        )
        compound_interaction.exec.time = atomic_interactions[-1].exec.time

    @staticmethod
    def _set_itb_test_case_status(itb_test_case: TestCaseDetails, robot_status: str):
        robot_status = robot_status.lower()
        if robot_status == "pass":
            itb_test_case.exec.status = ActivityStatus.Performed
            itb_test_case.exec.verdict = ExecutionVerdict.Pass
        elif robot_status == "fail":
            itb_test_case.exec.status = ActivityStatus.Performed
            itb_test_case.exec.verdict = ExecutionVerdict.Fail
        else:
            itb_test_case.exec.status = ActivityStatus.Running
            itb_test_case.exec.verdict = ExecutionVerdict.Undefined

    def end_suite(self, suite: TestSuite):
        if not suite.metadata.get("uniqueID") or len(suite.suites):
            return
        test_case_set = self.json_reader.read_test_case_set(suite.metadata["uniqueID"])
        if not test_case_set:
            return
        test_case_set.exec.verdict = suite.status

        for testcase in test_case_set.testCases:
            current_itb_test_case = self.itb_test_case_catalog.get(testcase.uniqueID)
            if current_itb_test_case is None:
                continue
            testcase.exec.verdict = current_itb_test_case.exec.verdict
            testcase.exec.status = current_itb_test_case.exec.status
            testcase.exec.execStatus = current_itb_test_case.exec.execStatus
            testcase.exec.comments = current_itb_test_case.exec.comments
        suite_start_time = "99999999 00:00:00.000"
        suite_end_time = "00000000 00:00:00.000"
        table_content = []
        for test in suite.tests:
            suite_start_time = min(suite_start_time, test.starttime)
            suite_end_time = max(suite_end_time, test.endtime)
            test_chain = get_test_chain(test.name, self.phase_pattern)

            if test_chain:
                name = test_chain.name if test_chain.index == 1 else ""
                phase = f"Phase {test_chain.index}/{test_chain.length}"
            else:
                name = test.name
                phase = ""
            if test.status != "PASS":
                message = re.sub(r"\s*itb-reference:\s*(\S*)", "", test.message)
                message = (
                    message[len("*HTML*") :]
                    .replace('<hr>', '<br />')
                    .replace('<hr>', '<br />')
                    .strip()
                    if message.startswith("*HTML*")
                    else message
                )
            else:
                message = self.get_isotime_from_robot_timestamp(test.endtime)

            table_content.append(
                f"<td>{name}</td>"
                f"<td>{phase}</td>"
                f"<td {self.render_status(test.status)}><b>{test.status}</b></td>"
                f"<td><pre>{message}</pre></td>"
            )

        test_case_set.exec.comments = (
            "<html>"
            "<head>"
            "<style>"
            "td {padding: 5px; border: none; white-space: pre-wrap;} "
            "table {font-family: monospace; border: none;}"
            "</style>"
            "</head>"
            "<body>"
            "<pre>"
            f"Start Time:   {self.get_isotime_from_robot_timestamp(suite_start_time)}\n"
            f"End Time:     {self.get_isotime_from_robot_timestamp(suite_end_time)}\n"
            "</pre>"
            "<table>"
            "<tr>"
            f"{'</tr><tr>'.join(table_content)}"
            "</tr>"
            "</table>"
            "</body>"
            "</html>"
        )
        write_test_structure_element(self.json_result, test_case_set)
        logger.debug(
            f"Successfully wrote the result from suite "
            f"{test_case_set.uniqueID} to TestBench's Json Report."
        )

    @staticmethod
    def render_status(status):
        return (
            "style='"
            f"background-color: {BACKGROUND_COLOR.get(status, '#fff')}; "
            f"color: {COLOR.get(status, '#000')};'"
        )

    def end_result(self, result):
        tt_tree = self.json_reader.read_test_theme_tree()
        if tt_tree:
            test_suite_counter = 0
            for tse in tt_tree.nodes:
                if self.test_suites.get(tse.baseInformation.uniqueID) is None:
                    continue
                execution_result = self._get_execution_result(
                    self.test_suites[tse.baseInformation.uniqueID].status
                )
                tse.execution.verdict = execution_result["execution_verdict"]
                tse.execution.status = execution_result["activity_status"]
                test_suite_counter += 1
            write_test_structure_element(self.json_result, tt_tree)
            if test_suite_counter and self.itb_test_case_catalog:
                logger.info(f"Successfully read {test_suite_counter} test suites.")
            else:
                logger.warning("No test suites with execution information found.")
            if self.create_zip:
                directory_to_zip(self.json_result, self.json_result_path)
            elif self.json_result != self.json_result_path:
                copytree(self.json_dir, self.json_result_path, dirs_exist_ok=True)
                copytree(self.json_result, self.json_result_path, dirs_exist_ok=True)
            self.tempdir.cleanup()
        logger.info("Successfully wrote the robot execution results to TestBench's Json Report.")

    @staticmethod
    def _get_execution_result(robot_status: str) -> Dict:
        robot_status = robot_status.lower()
        if robot_status == "pass":
            return {
                "execution_verdict": ExecutionVerdict.Pass,
                "activity_status": ActivityStatus.Performed,
            }
        if robot_status == "fail":
            return {
                "execution_verdict": ExecutionVerdict.Fail,
                "activity_status": ActivityStatus.Performed,
            }
        return {
            "execution_verdict": ExecutionVerdict.Undefined,
            "activity_status": ActivityStatus.Skipped,
        }

    @staticmethod
    def _get_interaction_result(robot_status: str) -> InteractionVerdict:
        robot_status = robot_status.upper()
        if robot_status == "PASS":
            return InteractionVerdict.Pass
        if robot_status == "FAIL":
            return InteractionVerdict.Fail
        if robot_status == "NOT RUN":
            return InteractionVerdict.Undefined
        return InteractionVerdict.Skipped


class TestChain:
    def __init__(self, name, index, length):
        self.name = str(name)
        self.index = int(index)
        self.length = int(length)


def get_test_chain(test_name: str, phase_pattern: str) -> Optional[TestChain]:
    matcher = re.match(get_test_chain_pattern(phase_pattern), test_name)
    if matcher:
        return TestChain(*matcher.groups())
    return None


def get_test_chain_pattern(phase_pattern: str) -> str:
    testcase_placeholder = str(uuid.uuid4().int)
    index_placeholder = str(uuid.uuid4().int)
    length_placeholder = str(uuid.uuid4().int)
    raw_pattern = re.escape(
        phase_pattern.format(
            testcase=testcase_placeholder,
            index=index_placeholder,
            length=length_placeholder,
        )
    )
    return (
        raw_pattern.replace(testcase_placeholder, r"(.+?)")
        .replace(index_placeholder, r"(\d)")
        .replace(length_placeholder, r"(\d)")
    )


def get_normalized_keyword_name(keyword_name: str) -> str:
    return re.sub(r"[\s_]", "", keyword_name.lower())


def is_normalized_equal(kw_one: str, kw_two: str) -> bool:
    norm_kw_one = get_normalized_keyword_name(kw_one)
    norm_kw_two = get_normalized_keyword_name(kw_two)
    return norm_kw_one == norm_kw_two



import os
import sys
from typing import Dict

from robot.api import ExecutionResult

from .config import Configuration
from .log import logger, setup_logger
from .result_writer import ResultWriter


def robot2testbench(
    json_input_report: str,
    robot_result_xml: str,
    json_output_result: str = None,
    config: Dict = None,
):
    if not os.path.exists(json_input_report):
        sys.exit("Could not find json directory or zip file at the given path.")
    if not os.path.exists(robot_result_xml):
        sys.exit("Robot result xml does not exist at the given path.")
    configuration = Configuration.from_dict(config)
    setup_logger(configuration)
    logger.debug("Config file loaded.")
    result = ExecutionResult(robot_result_xml)
    logger.debug("Robot framework result xml loaded.")
    result.visit(
        ResultWriter(json_input_report, json_output_result, configuration, robot_result_xml)
    )



from typing import Dict

from .config import Configuration
from .json_reader import TestBenchJsonReader
from .log import logger, setup_logger

# from .robot_run import RobotSuiteRunner
from .testbench2rf import create_test_suites
from .testsuite_write import write_test_suites
from .utils import PathResolver, get_directory


def testbench2robotframework(json_report: str, config: Dict):
    configuration = Configuration.from_dict(config)
    setup_logger(configuration)
    logger.debug("Config file loaded.")
    json_report = get_directory(json_report)
    reader = TestBenchJsonReader(json_report)
    path_resolver = PathResolver(
        reader.test_theme_tree,
        tuple(reader.get_test_case_set_catalog().keys()),
        configuration.logSuiteNumbering,
    )
    test_suites = create_test_suites(
        reader.get_test_case_set_catalog(), path_resolver, configuration
    )
    # suite_runner = RobotSuiteRunner(test_suites, path_resolver)
    # suite_runner.run_suites()
    if not test_suites:
        logger.warning("There are no test suites in the exported TestBench Projekt.")
        return
    write_test_suites(test_suites, configuration)



import json
import sys
from dataclasses import dataclass
from json import JSONDecodeError
from os import path
from typing import Dict, List, Optional

from .log import logger
from .model import TestCaseDetails, TestCaseSetDetails, TestStructureTree, TestStructureTreeNodeType

TEST_STRUCTURE_TREE_FILE = "TestThemeTree.json"


@dataclass
class TestCaseSet:
    details: TestCaseSetDetails
    test_cases: Dict[str, TestCaseDetails]

    @property
    def metadata(self) -> Dict[str, str]:
        return {
            "UniqueID": self.details.uniqueID,
            "Name": self.details.name,
            "Numbering": self.details.numbering,
        }


class TestBenchJsonReader:
    def __init__(self, json_dir):
        self.json_dir = json_dir
        self._test_theme_tree: Optional[TestStructureTree] = None
        self._test_case_sets: Dict[str, TestCaseSetDetails] = {}
        self._test_cases: Dict[str, TestCaseDetails] = {}
        if not json_dir:
            logger.warning("No jsonReport path given.")
            sys.exit()

    @property
    def test_theme_tree(self) -> TestStructureTree:
        if not self._test_theme_tree:
            test_theme_path = path.join(self.json_dir, TEST_STRUCTURE_TREE_FILE)
            logger.debug(f"Loading TestThemeTree from {test_theme_path}")
            test_structure_tree = read_json(test_theme_path)
            self._test_theme_tree = TestStructureTree.from_dict(test_structure_tree)
            logger.info(f"{len(self._test_theme_tree.nodes)} nodes from TestThemeTree loaded.")
        return self._test_theme_tree

    @property
    def test_case_sets(self) -> Dict[str, TestCaseSetDetails]:
        if not self._test_case_sets:
            for tcs_uid in self.get_test_case_set_uids():
                test_case_set = self.read_test_case_set(tcs_uid)
                if test_case_set is not None:
                    self._test_case_sets[tcs_uid] = test_case_set
                    logger.debug(f"TestCaseSetDetails {tcs_uid} loaded.")
                else:
                    logger.debug(f"TestCaseSetDetails {tcs_uid} not found.")
            logger.info(f"{len(self._test_case_sets)} TestCaseSetDetails loaded.")
        return self._test_case_sets

    @property
    def test_cases(self) -> Dict[str, TestCaseDetails]:
        if not self._test_cases:
            for tcs_uid in self.test_case_sets:
                tc_uids = self.get_test_case_uids(tcs_uid)
                self._read_test_cases(tc_uids)
        return self._test_cases

    def _read_test_cases(self, tc_uids):
        for tc_uid in tc_uids:
            test_case = self.read_test_case(tc_uid)
            if test_case is not None:
                self._test_cases[tc_uid] = test_case
                logger.debug(f"TestCaseDetails {tc_uid} loaded.")

    def get_test_case_set_catalog(self):
        tcs_catalog: Dict[str, TestCaseSet] = {}
        for tcs_uid, tcs in self.test_case_sets.items():
            tc_catalog: Dict[str, TestCaseDetails] = {}
            for tc_uid in self.get_test_case_uids(tcs_uid):
                tc_catalog[tc_uid] = self.test_cases[tc_uid]
            tcs_catalog[tcs_uid] = TestCaseSet(tcs, tc_catalog)
        return tcs_catalog

    def get_test_case_set_uids(self) -> List[str]:
        nodes = self.test_theme_tree.nodes
        return [
            tse.baseInformation.uniqueID
            for tse in nodes
            if tse.elementType == TestStructureTreeNodeType.TestCaseSet
        ]

    def get_test_case_uids(self, test_case_set_uid: str) -> List[str]:
        if not self._test_case_sets:
            test_case_set = self.read_test_case_set(test_case_set_uid)
            if test_case_set is None:
                logger.debug(f"TestCaseSet with uid '{test_case_set_uid}' not found.")
                return []
        else:
            test_case_set = self.test_case_sets[test_case_set_uid]
        return [tc.uniqueID for tc in test_case_set.testCases]

    def read_test_case_set(self, uid) -> Optional[TestCaseSetDetails]:
        tcs_dict = read_json(path.join(self.json_dir, f"{uid}.json"))
        if tcs_dict is None:
            return None
        return TestCaseSetDetails.from_dict(tcs_dict)

    def read_test_case(self, uid) -> Optional[TestCaseDetails]:
        tc_dict = read_json(path.join(self.json_dir, f"{uid}.json"))
        if tc_dict is None:
            return None
        return TestCaseDetails.from_dict(tc_dict)

    def read_test_theme_tree(self) -> Optional[TestStructureTree]:
        test_structure_tree = read_json(path.join(self.json_dir, TEST_STRUCTURE_TREE_FILE))
        if test_structure_tree is None:
            return None
        return TestStructureTree.from_dict(test_structure_tree)


def read_json(filepath: str):  # ToDo Configure to run silent or raise
    try:
        with open(filepath, encoding='utf-8') as json_file:
            return json.load(json_file)
    except FileNotFoundError:
        logger.debug(f"Cannot find json file {filepath}:")
        return None
    except JSONDecodeError as error:  # pylint: disable=broad-except
        logger.warning(f"Cannot decode json file {filepath}:")
        logger.warning(error)
        return None



import pytest

from testbench2robotframework.utils import replace_invalid_characters

invalid_chars = ['<', '>', ':', '"', '/', '\\', '|', '?', '*']


@pytest.mark.parametrize("invalid_char", invalid_chars)
def test_filename_does_not_contain_invalid_char(invalid_char):
    filename = f"1.2.1 Testsuite with {invalid_char} in the name.robot"
    assert replace_invalid_characters(filename).find(invalid_char) == -1



# Copyright 2022-     imbus AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from os import path

import robot

from testbench2robotframework import __version__

from .config import write_default_config
from .json_reader import read_json
from .robotframework2testbench import robot2testbench
from .testbench2robotframework import testbench2robotframework
from .utils import arg_parser


def run():
    args = arg_parser.parse_args()
    if args.subcommand is None and not args.version:
        arg_parser.print_help()
        sys.exit()
    if args.version:
        print_version()
        sys.exit()
    if not path.isfile(args.config):
        write_default_config(args.config)
    configuration = read_json(args.config)
    if args.subcommand == 'write':
        testbench2robotframework(args.jsonReport[0], configuration)
    elif args.subcommand == 'read':
        robot2testbench(args.jsonReport[0], args.output, args.result, configuration)


def print_version():
    print(
        f'TestBench2RobotFramework {__version__} with '
        f'[Robot Framework {robot.version.get_full_version()}]'
    )


if __name__ == "__main__":
    run()
